/*
 *  libxcpc by Davide Libenzi (Exception and resource handling in C)
 *  Copyright (C) 2002  Davide Libenzi
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Davide Libenzi <davidel@xmailserver.org>
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>
#include "libxcpc.h"



static void xcpc_test__exit(int err);
static void xcpc_test__free1(void *data);
static int xcpc_test_throw(xcpc_ctx ctx, int n, void *data);
static void xcpc_test_1(xcpc_ctx ctx, int n);
static void xcpc_test_2(xcpc_ctx ctx, int n);
static void xcpc_test_3(xcpc_ctx ctx, int n);
static void xcpc_test_4(xcpc_ctx ctx, int n);
static void xcpc_test_5(xcpc_ctx ctx, int n);
static void xcpc_test_6(xcpc_ctx ctx, int n);
static void xcpc_test_7(xcpc_ctx ctx, int n);
static void xcpc_test_8(xcpc_ctx ctx);
static void xcpc_test_run(xcpc_ctx ctx);



static jmp_buf exctx;



static void xcpc_test__exit(int err) {

	fprintf(stderr, "ahiee! exception %d has been thrown but has not been caught\n",
		err);
	longjmp(exctx, err);
}

static void xcpc_test__free1(void *data) {

	fprintf(stdout, "Freeing string: '%s'\n", data);
}

static int xcpc_test_throw(xcpc_ctx ctx, int n, void *data) {

	XCPC_THROW(ctx, n, data);

	return 0;
}

static void xcpc_test_1(xcpc_ctx ctx, int n) {

	fprintf(stdout, "trying to throw exception %d (local catch)\n", n);
	XCPC_TRY(ctx) {
		xcpc_resource_add(ctx, (void *) "xcpc_test_1",
				  xcpc_test__free1);

		xcpc_test_throw(ctx, n, NULL);
	}
	XCPC_CATCH(n) {
		fprintf(stdout, "local: caught exception %d\n", n);
	}
	XCPC_END_TRY;
}

static void xcpc_test_2(xcpc_ctx ctx, int n) {

	fprintf(stdout, "trying to throw exception %d (no local catch)\n", n);
	XCPC_TRY(ctx) {
		xcpc_resource_add(ctx, (void *) "xcpc_test_2",
				  xcpc_test__free1);

		xcpc_test_throw(ctx, n, NULL);
	}
	XCPC_CATCH(n + 1) {
		fprintf(stdout, "local: (unlikely) caught exception %d\n", n + 1);
	}
	XCPC_END_TRY;
}

static void xcpc_test_3(xcpc_ctx ctx, int n) {

	fprintf(stdout, "trying to throw exception %d (local catch and re-throw)\n", n);
	XCPC_TRY(ctx) {
		xcpc_resource_add(ctx, (void *) "xcpc_test_3",
				  xcpc_test__free1);

		xcpc_test_throw(ctx, n, NULL);
	}
	XCPC_CATCH(n) {
		fprintf(stdout, "local: caught exception %d. re-throwing ...\n",
			n);
		XCPC_RETHROW(ctx);
	}
	XCPC_END_TRY;
}

static void xcpc_test_4(xcpc_ctx ctx, int n) {

	fprintf(stdout, "trying to throw exception %d (no local catch - simply TRY)\n", n);
	XCPC_TRY(ctx) {
		xcpc_resource_add(ctx, (void *) "xcpc_test_4",
				  xcpc_test__free1);

		xcpc_test_throw(ctx, n, NULL);
	}
	XCPC_END_TRY;
}

static void xcpc_test_5(xcpc_ctx ctx, int n) {

	fprintf(stdout, "trying to throw exception %d (no local catch - with DATA)\n", n);
	XCPC_TRY(ctx) {
		char *str;
		xcpc_resource_add(ctx, (void *) "xcpc_test_5",
				  xcpc_test__free1);

		str = xcpc__malloc(ctx, NULL, 128);
		snprintf(str, 128, "data for exception %d", n);

		xcpc_test_throw(ctx, n, str);
	}
	XCPC_END_TRY;
}

static void xcpc_test_6(xcpc_ctx ctx, int n) {
	xcpc_ctx sctx;

	if ((sctx = xcpc_context_create(ctx)) == NULL) {
		fprintf(stderr, "unable to create context\n");
		return 1;
	}
	fprintf(stdout, "trying to throw exception %d (no local catch - with DATA and subctx)\n", n);
	XCPC_TRY(sctx) {
		char *str;
		xcpc_resource_add(sctx, (void *) "xcpc_test_6",
				  xcpc_test__free1);

		str = xcpc__malloc(sctx, NULL, 128);
		snprintf(str, 128, "data for exception %d", n);

		xcpc_test_throw(sctx, n, str);
	}
	XCPC_END_TRY;

	xcpc_context_free(sctx);
}

static void xcpc_test_7(xcpc_ctx ctx, int n) {

	fprintf(stdout, "trying to throw exception %d (local catch and throw a new one)\n", n);
	XCPC_TRY(ctx) {
		char *str;
		xcpc_resource_add(ctx, (void *) "xcpc_test_7",
				  xcpc_test__free1);

		str = xcpc__malloc(ctx, NULL, 128);
		snprintf(str, 128, "data for exception %d", n);

		xcpc_test_throw(ctx, n, str);
	}
	XCPC_CATCH(n) {
		fprintf(stdout, "local: caught exception %d. throwing %d (same data) ...\n",
			n, n + 1);
		XCPC_THROW(ctx, n + 1, XCPC_EXCEPT_DATA);
	}
	XCPC_END_TRY;
}

static void xcpc_test_8(xcpc_ctx ctx) {
	char const *path = "/boogao/moocao";

	fprintf(stdout, "trying to open a bogus file '%s' (local catch)\n", path);
	XCPC_TRY(ctx) {
		xcpc_resource_add(ctx, (void *) "xcpc_test_8",
				  xcpc_test__free1);

		xcpc__fopen(ctx, NULL, path, "r");
	}
	XCPC_CATCH_ANY {
		fprintf(stdout, "local: caught exception %d (path = '%s')\n",
			XCPC_EXCEPTION, XCPC_EXCEPT_DATA);
	}
	XCPC_END_TRY;
}

static void xcpc_test_run(xcpc_ctx ctx) {

	/* xcpc_test_1 */
	XCPC_TRY(ctx) {
		xcpc_test_1(ctx, 1);
	}
	XCPC_CATCH_ANY {
		fprintf(stdout, "global: caught exception %d\n", XCPC_EXCEPTION);
	}
	XCPC_END_TRY;


	/* xcpc_test_2 */
	XCPC_TRY(ctx) {
		xcpc_test_2(ctx, 2);
	}
	XCPC_CATCH_ANY {
		fprintf(stdout, "global: caught exception %d\n", XCPC_EXCEPTION);
	}
	XCPC_END_TRY;


	/* xcpc_test_3 */
	XCPC_TRY(ctx) {
		xcpc_test_3(ctx, 3);
	}
	XCPC_CATCH_ANY {
		fprintf(stdout, "global: caught exception %d\n", XCPC_EXCEPTION);
	}
	XCPC_END_TRY;

	/* xcpc_test_4 */
	XCPC_TRY(ctx) {
		xcpc_test_4(ctx, 4);
	}
	XCPC_CATCH(4) {
		fprintf(stdout, "global: caught exception %d\n", XCPC_EXCEPTION);
	}
	XCPC_END_TRY;


	/* xcpc_test_5 */
	XCPC_TRY(ctx) {
		xcpc_test_5(ctx, 5);
	}
	XCPC_CATCH_ANY {
		fprintf(stdout, "global: caught exception %d (data = '%s')\n",
			XCPC_EXCEPTION, XCPC_EXCEPT_DATA);
	}
	XCPC_END_TRY;


	/* xcpc_test_6 */
	XCPC_TRY(ctx) {
		xcpc_test_6(ctx, 6);
	}
	XCPC_CATCH_ANY {
		fprintf(stdout, "global: caught exception %d (data = '%s')\n",
			XCPC_EXCEPTION, XCPC_EXCEPT_DATA);
	}
	XCPC_END_TRY;


	/* xcpc_test_7 */
	XCPC_TRY(ctx) {
		xcpc_test_7(ctx, 7);
	}
	XCPC_CATCH_ANY {
		fprintf(stdout, "global: caught exception %d (data = '%s')\n",
			XCPC_EXCEPTION, XCPC_EXCEPT_DATA);
	}
	XCPC_END_TRY;

	/* xcpc_test_8 */
	xcpc_test_8(ctx);
}

int main(int ac, char **av) {
	xcpc_ctx ctx;

	xcpc_set_exitproc(xcpc_test__exit);
	if (setjmp(exctx) == 0) {
		ctx = xcpc_context_create(NULL);

		XCPC_TRY(ctx) {
			xcpc_resource_add(ctx, (void *) "main1",
					  xcpc_test__free1);
			xcpc_resource_add(ctx, (void *) "main2",
					  xcpc_test__free1);

			xcpc_test_run(ctx);
		}
		XCPC_CATCH_ANY {
			fprintf(stdout, "main: caught exception %d (data = %p)\n",
				XCPC_EXCEPTION, XCPC_EXCEPT_DATA);
		}
		XCPC_END_TRY;

		xcpc_context_free(ctx);
	}

	return 0;
}

